import logging
from collections import defaultdict
from urllib.parse import quote

from opentelemetry.proto.collector.trace.v1.trace_service_pb2 import ExportTraceServiceRequest

from mlflow.entities import Assessment, Span, Trace, TraceInfo, TraceLocation
from mlflow.entities.assessment import ExpectationValue, FeedbackValue
from mlflow.entities.trace_location import UCSchemaLocation as UCSchemaLocationEntity
from mlflow.environment_variables import (
    MLFLOW_ASYNC_TRACE_LOGGING_RETRY_TIMEOUT,
    MLFLOW_TRACING_SQL_WAREHOUSE_ID,
)
from mlflow.exceptions import MlflowException, RestException
from mlflow.protos.databricks_pb2 import (
    ALREADY_EXISTS,
    BAD_REQUEST,
    ENDPOINT_NOT_FOUND,
    INVALID_PARAMETER_VALUE,
    ErrorCode,
)
from mlflow.protos.databricks_tracing_pb2 import Assessment as ProtoAssessment
from mlflow.protos.databricks_tracing_pb2 import (
    BatchGetTraces,
    BatchLinkTraceToRun,
    BatchUnlinkTraceFromRun,
    CreateAssessment,
    CreateTraceInfo,
    CreateTraceUCStorageLocation,
    DatabricksTrackingService,
    DeleteAssessment,
    DeleteTraceTag,
    GetAssessment,
    GetTraceInfo,
    LinkExperimentToUCTraceLocation,
    SearchTraces,
    SetTraceTag,
    UnLinkExperimentToUCTraceLocation,
    UpdateAssessment,
)
from mlflow.protos.databricks_tracing_pb2 import TraceInfo as ProtoTraceInfo
from mlflow.protos.service_pb2 import GetOnlineTraceDetails, MlflowService, SearchUnifiedTraces
from mlflow.store.tracking import SEARCH_TRACES_DEFAULT_MAX_RESULTS
from mlflow.store.tracking.rest_store import RestStore
from mlflow.tracing.utils import parse_trace_id_v4
from mlflow.tracing.utils.otlp import OTLP_TRACES_PATH
from mlflow.utils.databricks_tracing_utils import (
    assessment_to_proto,
    trace_from_proto,
    trace_location_to_proto,
    uc_schema_location_from_proto,
    uc_schema_location_to_proto,
)
from mlflow.utils.databricks_utils import get_databricks_workspace_client_config
from mlflow.utils.proto_json_utils import message_to_json
from mlflow.utils.rest_utils import (
    _REST_API_PATH_PREFIX,
    _V4_REST_API_PATH_PREFIX,
    _V4_TRACE_REST_API_PATH_PREFIX,
    extract_api_info_for_service,
    get_single_assessment_endpoint_v4,
    get_single_trace_endpoint_v4,
    http_request,
    verify_rest_response,
)

DATABRICKS_UC_TABLE_HEADER = "X-Databricks-UC-Table-Name"

_logger = logging.getLogger(__name__)


class DatabricksTracingRestStore(RestStore):
    """
    Client for a databricks tracking server accessed via REST API calls.
    This is only used for Databricks-specific tracing APIs, all other APIs including
    runs, experiments, models etc. should be implemented in the RestStore.

    Args
        get_host_creds: Method to be invoked prior to every REST request to get the
            :py:class:`mlflow.rest_utils.MlflowHostCreds` for the request. Note that this
            is a function so that we can obtain fresh credentials in the case of expiry.
    """

    _METHOD_TO_INFO = extract_api_info_for_service(
        MlflowService, _REST_API_PATH_PREFIX
    ) | extract_api_info_for_service(DatabricksTrackingService, _V4_REST_API_PATH_PREFIX)

    def __init__(self, get_host_creds):
        super().__init__(get_host_creds)

    def _call_endpoint(
        self,
        api,
        json_body=None,
        endpoint=None,
        retry_timeout_seconds=None,
        response_proto=None,
    ):
        try:
            return super()._call_endpoint(
                api,
                json_body=json_body,
                endpoint=endpoint,
                retry_timeout_seconds=retry_timeout_seconds,
                response_proto=response_proto,
            )
        except RestException as e:
            if (
                e.error_code == ErrorCode.Name(INVALID_PARAMETER_VALUE)
                and "Could not resolve a SQL warehouse ID" in e.message
            ):
                raise MlflowException(
                    message=(
                        "SQL warehouse ID is required for accessing traces in UC tables.\n"
                        f"Please set the {MLFLOW_TRACING_SQL_WAREHOUSE_ID.name} environment "
                        "variable to your SQL warehouse ID.\n"
                        "```\nexport MLFLOW_TRACING_SQL_WAREHOUSE_ID=<your_sql_warehouse_id>\n```\n"
                        "See https://docs.databricks.com/compute/sql-warehouse for how to "
                        "set up a SQL warehouse and get its ID."
                    ),
                    error_code=BAD_REQUEST,
                ) from e
            raise

    def start_trace(self, trace_info: TraceInfo) -> TraceInfo:
        """
        Create a new trace using the V4 API format.

        Args:
            trace_info: The TraceInfo object to create in the backend. Currently, this
                only supports trace_location with uc_schema, or mlflow_experiment that's
                linked to a UC table.

        Returns:
            The returned TraceInfo object from the backend.
        """
        try:
            if trace_info._is_v4():
                return self._start_trace_v4(trace_info)

        # Temporarily we capture all exceptions and fallback to v3 if the trace location is not uc
        # TODO: remove this once the endpoint is fully rolled out
        except Exception as e:
            if trace_info.trace_location.mlflow_experiment is None:
                _logger.debug("MLflow experiment is not set for trace, cannot fallback to V3 API.")
                raise
            _logger.debug(f"Falling back to V3 API due to {e!s}")
        return super().start_trace(trace_info)

    def _start_trace_v4(self, trace_info: TraceInfo) -> TraceInfo:
        location, otel_trace_id = parse_trace_id_v4(trace_info.trace_id)
        if location is None:
            raise MlflowException("Invalid trace ID format for v4 API.")

        req_body = message_to_json(trace_info.to_proto())
        response_proto = self._call_endpoint(
            CreateTraceInfo,
            req_body,
            endpoint=f"{_V4_REST_API_PATH_PREFIX}/mlflow/traces/{location}/{otel_trace_id}/info",
            retry_timeout_seconds=MLFLOW_ASYNC_TRACE_LOGGING_RETRY_TIMEOUT.get(),
            response_proto=ProtoTraceInfo(),
        )
        return TraceInfo.from_proto(response_proto)

    def batch_get_traces(self, trace_ids: list[str], location: str) -> list[Trace]:
        """
        Get a batch of complete traces with spans for given trace ids.

        Args:
            trace_ids: List of trace IDs to fetch.
            location: Location of the trace. For example, "catalog.schema" for UC schema.

        Returns:
            List of Trace objects.
        """
        trace_ids = [parse_trace_id_v4(trace_id)[1] for trace_id in trace_ids]
        req_body = message_to_json(
            BatchGetTraces(
                location_id=location,
                trace_ids=trace_ids,
                sql_warehouse_id=MLFLOW_TRACING_SQL_WAREHOUSE_ID.get(),
            )
        )
        response_proto = self._call_endpoint(
            BatchGetTraces,
            req_body,
            endpoint=f"{_V4_TRACE_REST_API_PATH_PREFIX}/{location}/batchGet",
        )
        return [trace_from_proto(proto, location) for proto in response_proto.traces]

    def get_trace_info(self, trace_id: str) -> TraceInfo:
        """
        Get the trace info matching the `trace_id`.

        Args:
            trace_id: String id of the trace to fetch.

        Returns:
            The fetched ``mlflow.entities.TraceInfo`` object.
        """
        location, trace_id = parse_trace_id_v4(trace_id)
        if location is not None:
            sql_warehouse_id = MLFLOW_TRACING_SQL_WAREHOUSE_ID.get()
            trace_v4_req_body = message_to_json(
                GetTraceInfo(
                    trace_id=trace_id, location=location, sql_warehouse_id=sql_warehouse_id
                )
            )
            endpoint = f"{get_single_trace_endpoint_v4(location, trace_id)}/info"
            response_proto = self._call_endpoint(GetTraceInfo, trace_v4_req_body, endpoint=endpoint)
            return TraceInfo.from_proto(response_proto.trace.trace_info)

        return super().get_trace_info(trace_id)

    def set_trace_tag(self, trace_id: str, key: str, value: str):
        """
        Set a tag on the trace with the given trace_id.

        Args:
            trace_id: The ID of the trace.
            key: The string key of the tag.
            value: The string value of the tag.
        """
        location, trace_id = parse_trace_id_v4(trace_id)
        if location is not None:
            endpoint = f"{get_single_trace_endpoint_v4(location, trace_id)}/tags"
            req_body = message_to_json(
                SetTraceTag(
                    key=key,
                    value=value,
                )
            )
            self._call_endpoint(SetTraceTag, req_body, endpoint=endpoint)
            return
        return super().set_trace_tag(trace_id, key, value)

    def delete_trace_tag(self, trace_id: str, key: str):
        """
        Delete a tag on the trace with the given trace_id.

        Args:
            trace_id: The ID of the trace.
            key: The string key of the tag.
        """
        location, trace_id = parse_trace_id_v4(trace_id)
        if location is not None:
            sql_warehouse_id = MLFLOW_TRACING_SQL_WAREHOUSE_ID.get()
            encoded_key = quote(key, safe="")
            endpoint = f"{get_single_trace_endpoint_v4(location, trace_id)}/tags/{encoded_key}"
            req_body = message_to_json(DeleteTraceTag(sql_warehouse_id=sql_warehouse_id))
            self._call_endpoint(DeleteTraceTag, req_body, endpoint=endpoint)
            return
        return super().delete_trace_tag(trace_id, key)

    def search_traces(
        self,
        experiment_ids: list[str] | None = None,
        filter_string: str | None = None,
        max_results: int = SEARCH_TRACES_DEFAULT_MAX_RESULTS,
        order_by: list[str] | None = None,
        page_token: str | None = None,
        model_id: str | None = None,
        locations: list[str] | None = None,
    ) -> tuple[list[TraceInfo], str | None]:
        # This API is not client-facing, so we should always use `locations`.
        if experiment_ids is not None:
            raise MlflowException("`experiment_ids` is deprecated, use `locations` instead.")
        if not locations:
            raise MlflowException.invalid_parameter_value(
                "At least one location must be specified for searching traces."
            )

        # model_id is only supported by V3 API
        if model_id is not None:
            return self._search_unified_traces(
                model_id=model_id,
                locations=locations,
                filter_string=filter_string,
                max_results=max_results,
                order_by=order_by,
                page_token=page_token,
            )

        contain_uc_schemas = False
        trace_locations = []
        for location in locations:
            match location.split("."):
                case [experiment_id]:
                    trace_locations.append(
                        trace_location_to_proto(TraceLocation.from_experiment_id(experiment_id))
                    )
                case [catalog, schema]:
                    trace_locations.append(
                        trace_location_to_proto(
                            TraceLocation.from_databricks_uc_schema(catalog, schema)
                        )
                    )
                    contain_uc_schemas = True
                case _:
                    raise MlflowException.invalid_parameter_value(
                        f"Invalid location type: {location}. Expected type: "
                        "`<catalog_name>.<schema_name>` or `<experiment_id>`."
                    )

        request = SearchTraces(
            locations=trace_locations,
            filter=filter_string,
            max_results=max_results,
            order_by=order_by,
            page_token=page_token,
            sql_warehouse_id=MLFLOW_TRACING_SQL_WAREHOUSE_ID.get(),
        )
        req_body = message_to_json(request)
        try:
            response_proto = self._call_endpoint(
                SearchTraces,
                req_body,
                endpoint=f"{_V4_TRACE_REST_API_PATH_PREFIX}/search",
            )
        except MlflowException as e:
            # There are 2 expected failure cases:
            # 1. Server does not support SearchTracesV4 API yet.
            # 2. Server supports V4 API but the experiment location is not supported yet.
            # For these known cases, MLflow fallback to V3 API.
            if e.error_code == ErrorCode.Name(ENDPOINT_NOT_FOUND):
                if contain_uc_schemas:
                    raise MlflowException.invalid_parameter_value(
                        "Searching traces in UC tables is not supported yet. Only experiment IDs "
                        "are supported for searching traces."
                    )
                _logger.debug("SearchTracesV4 API is not available yet. Falling back to V3 API.")
            elif (
                e.error_code == ErrorCode.Name(INVALID_PARAMETER_VALUE)
                and "locations not yet supported" in e.message
            ):
                if contain_uc_schemas:
                    raise MlflowException.invalid_parameter_value(
                        "The `locations` parameter cannot contain both MLflow experiment and UC "
                        "schema in the same request. Please specify only one type of location "
                        "at a time."
                    )
                _logger.debug("Experiment locations are not supported yet. Falling back to V3 API.")
            else:
                raise

            return self._search_traces(
                locations=locations,
                filter_string=filter_string,
                max_results=max_results,
                order_by=order_by,
                page_token=page_token,
            )

        trace_infos = [TraceInfo.from_proto(t) for t in response_proto.trace_infos]
        return trace_infos, response_proto.next_page_token or None

    def _search_unified_traces(
        self,
        model_id: str,
        locations: list[str],
        filter_string: str | None = None,
        max_results: int = SEARCH_TRACES_DEFAULT_MAX_RESULTS,
        order_by: list[str] | None = None,
        page_token: str | None = None,
    ) -> tuple[list[TraceInfo], str | None]:
        request = SearchUnifiedTraces(
            model_id=model_id,
            sql_warehouse_id=MLFLOW_TRACING_SQL_WAREHOUSE_ID.get(),
            experiment_ids=locations,
            filter=filter_string,
            max_results=max_results,
            order_by=order_by,
            page_token=page_token,
        )
        req_body = message_to_json(request)
        response_proto = self._call_endpoint(SearchUnifiedTraces, req_body)
        # Convert TraceInfo (v2) objects to TraceInfoV3 objects for consistency
        trace_infos = [TraceInfo.from_proto(t) for t in response_proto.traces]
        return trace_infos, response_proto.next_page_token or None

    def get_online_trace_details(
        self,
        trace_id: str,
        source_inference_table: str,
        source_databricks_request_id: str,
    ):
        req = GetOnlineTraceDetails(
            trace_id=trace_id,
            sql_warehouse_id=MLFLOW_TRACING_SQL_WAREHOUSE_ID.get(),
            source_inference_table=source_inference_table,
            source_databricks_request_id=source_databricks_request_id,
        )
        req_body = message_to_json(req)
        response_proto = self._call_endpoint(GetOnlineTraceDetails, req_body)
        return response_proto.trace_data

    def set_experiment_trace_location(
        self,
        location: UCSchemaLocationEntity,
        experiment_id: str,
        sql_warehouse_id: str | None = None,
    ) -> UCSchemaLocationEntity:
        req_body = message_to_json(
            CreateTraceUCStorageLocation(
                uc_schema=uc_schema_location_to_proto(location),
                sql_warehouse_id=sql_warehouse_id or MLFLOW_TRACING_SQL_WAREHOUSE_ID.get(),
            )
        )
        try:
            response = self._call_endpoint(
                CreateTraceUCStorageLocation,
                req_body,
                endpoint=f"{_V4_TRACE_REST_API_PATH_PREFIX}/location",
            )
            location = uc_schema_location_from_proto(response.uc_schema)
        except MlflowException as e:
            if e.error_code == ErrorCode.Name(ALREADY_EXISTS):
                _logger.debug(f"Trace UC storage location already exists: {location}")
            else:
                raise
        _logger.debug(f"Created trace UC storage location: {location}")

        # link experiment to uc trace location
        req_body = message_to_json(
            LinkExperimentToUCTraceLocation(
                experiment_id=experiment_id,
                uc_schema=uc_schema_location_to_proto(location),
            )
        )

        self._call_endpoint(
            LinkExperimentToUCTraceLocation,
            req_body,
            endpoint=f"{_V4_TRACE_REST_API_PATH_PREFIX}/{experiment_id}/link-location",
        )
        _logger.debug(f"Linked experiment {experiment_id} to UC trace location: {location}")
        return location

    def unset_experiment_trace_location(
        self, experiment_id: str, location: UCSchemaLocationEntity
    ) -> None:
        request = UnLinkExperimentToUCTraceLocation(
            experiment_id=experiment_id,
            uc_schema=uc_schema_location_to_proto(location),
        )
        endpoint = f"{_V4_TRACE_REST_API_PATH_PREFIX}/{experiment_id}/unlink-location"
        req_body = message_to_json(request)
        self._call_endpoint(
            UnLinkExperimentToUCTraceLocation,
            req_body,
            endpoint=endpoint,
        )
        _logger.debug(f"Unlinked experiment {experiment_id} from trace location: {location}")

    def log_spans(self, location: str, spans: list[Span], tracking_uri=None) -> list[Span]:
        if not spans:
            return []

        if tracking_uri is None:
            raise MlflowException(
                "`tracking_uri` must be provided to log spans to with Databricks tracking server."
            )

        endpoint = f"/api/2.0/otel{OTLP_TRACES_PATH}"
        try:
            config = get_databricks_workspace_client_config(tracking_uri)
        except Exception as e:
            raise MlflowException(
                "Failed to log spans to UC table: could not identify Databricks workspace "
                "configuration"
            ) from e

        request = ExportTraceServiceRequest()
        resource_spans = request.resource_spans.add()
        scope_spans = resource_spans.scope_spans.add()
        scope_spans.spans.extend(span.to_otel_proto() for span in spans)

        response = http_request(
            host_creds=self.get_host_creds(),
            endpoint=endpoint,
            method="POST",
            data=request.SerializeToString(),
            extra_headers={
                "Content-Type": "application/x-protobuf",
                DATABRICKS_UC_TABLE_HEADER: location,
                **config.authenticate(),
            },
        )
        verify_rest_response(response, endpoint)
        return spans

    def create_assessment(self, assessment: Assessment) -> Assessment:
        """
        Create an assessment entity in the backend store.

        Args:
            assessment: The assessment to log (without an assessment_id).

        Returns:
            The created Assessment object.
        """
        location, trace_id = parse_trace_id_v4(assessment.trace_id)
        if location is not None:
            req_body = message_to_json(assessment_to_proto(assessment))
            endpoint = self._append_sql_warehouse_id_param(
                f"{get_single_trace_endpoint_v4(location, trace_id)}/assessments",
            )
            response_proto = self._call_endpoint(
                CreateAssessment,
                req_body,
                endpoint=endpoint,
                response_proto=ProtoAssessment(),
            )
            return Assessment.from_proto(response_proto)

        return super().create_assessment(assessment)

    def update_assessment(
        self,
        trace_id: str,
        assessment_id: str,
        name: str | None = None,
        expectation: ExpectationValue | None = None,
        feedback: FeedbackValue | None = None,
        rationale: str | None = None,
        metadata: dict[str, str] | None = None,
    ) -> Assessment:
        """
        Update an existing assessment entity in the backend store.

        Args:
            trace_id: The ID of the trace.
            assessment_id: The ID of the assessment to update.
            name: The updated name of the assessment.
            expectation: The updated expectation value of the assessment.
            feedback: The updated feedback value of the assessment.
            rationale: The updated rationale of the feedback. Not applicable for expectations.
            metadata: Additional metadata for the assessment.
        """
        if expectation is not None and feedback is not None:
            raise MlflowException.invalid_parameter_value(
                "Exactly one of `expectation` or `feedback` should be specified."
            )

        location, parsed_trace_id = parse_trace_id_v4(trace_id)
        if location is not None:
            assessment = UpdateAssessment().assessment
            assessment.assessment_id = assessment_id
            catalog, schema = location.split(".")
            assessment.trace_location.CopyFrom(
                trace_location_to_proto(TraceLocation.from_databricks_uc_schema(catalog, schema)),
            )
            assessment.trace_id = parsed_trace_id
            # Field mask specifies which fields to update.
            mask = UpdateAssessment().update_mask
            if name is not None:
                assessment.assessment_name = name
                mask.paths.append("assessment_name")
            if expectation is not None:
                assessment.expectation.CopyFrom(expectation.to_proto())
                mask.paths.append("expectation")
            if feedback is not None:
                assessment.feedback.CopyFrom(feedback.to_proto())
                mask.paths.append("feedback")
            if rationale is not None:
                assessment.rationale = rationale
                mask.paths.append("rationale")
            if metadata is not None:
                assessment.metadata.update(metadata)
                mask.paths.append("metadata")

            endpoint = get_single_assessment_endpoint_v4(location, parsed_trace_id, assessment_id)
            endpoint = self._append_sql_warehouse_id_param(endpoint)

            if mask.paths:
                mask_param = ",".join(mask.paths)
                endpoint = f"{endpoint}&update_mask={mask_param}"

            req_body = message_to_json(assessment)
            response_proto = self._call_endpoint(
                UpdateAssessment,
                req_body,
                endpoint=endpoint,
                response_proto=ProtoAssessment(),
            )
            return Assessment.from_proto(response_proto)
        else:
            return super().update_assessment(
                trace_id, assessment_id, name, expectation, feedback, rationale, metadata
            )

    def get_assessment(self, trace_id: str, assessment_id: str) -> Assessment:
        """
        Get an assessment entity from the backend store.
        """

        location, trace_id = parse_trace_id_v4(trace_id)
        if location is not None:
            endpoint = self._append_sql_warehouse_id_param(
                get_single_assessment_endpoint_v4(location, trace_id, assessment_id)
            )
            response_proto = self._call_endpoint(
                GetAssessment, endpoint=endpoint, response_proto=ProtoAssessment()
            )
            return Assessment.from_proto(response_proto)

        return super().get_assessment(trace_id, assessment_id)

    def delete_assessment(self, trace_id: str, assessment_id: str):
        """
        Delete an assessment associated with a trace.

        Args:
            trace_id: String ID of the trace.
            assessment_id: String ID of the assessment to delete.
        """
        location, trace_id = parse_trace_id_v4(trace_id)
        if location is not None:
            endpoint = self._append_sql_warehouse_id_param(
                get_single_assessment_endpoint_v4(location, trace_id, assessment_id)
            )
            self._call_endpoint(DeleteAssessment, endpoint=endpoint)
        else:
            return super().delete_assessment(trace_id, assessment_id)

    def _group_traces_by_location(self, trace_ids: list[str]) -> dict[str | None, list[str]]:
        """
        Group trace IDs by location to separate V3 and V4 traces.

        Args:
            trace_ids: List of trace IDs (can be V3 or V4 format).

        Returns:
            Dict mapping location to list of trace IDs where:
            - None key: List of V3 trace IDs (without location prefix)
            - str keys: Location IDs (e.g., "catalog.schema") mapping to OTEL trace IDs
        """
        traces_by_location: dict[str | None, list[str]] = defaultdict(list)

        for trace_id in trace_ids:
            location_id, trace_id = parse_trace_id_v4(trace_id)
            traces_by_location[location_id].append(trace_id)

        return traces_by_location

    def _batch_link_traces_to_run(
        self, location_id: str, otel_trace_ids: list[str], run_id: str
    ) -> None:
        """
        Link multiple traces to a run by creating internal trace-to-run relationships.

        Args:
            location_id: The location ID (e.g., "catalog.schema") for the traces.
            otel_trace_ids: List of OTEL trace IDs to link to the run.
            run_id: ID of the run to link traces to.
        """
        if not otel_trace_ids:
            return

        req_body = message_to_json(
            BatchLinkTraceToRun(
                location_id=location_id,
                trace_ids=otel_trace_ids,
                run_id=run_id,
            )
        )
        endpoint = f"{_V4_TRACE_REST_API_PATH_PREFIX}/{location_id}/link-to-run/batchCreate"
        self._call_endpoint(BatchLinkTraceToRun, req_body, endpoint=endpoint)

    def _batch_unlink_traces_from_run(
        self, location_id: str, otel_trace_ids: list[str], run_id: str
    ) -> None:
        """
        Unlink multiple traces from a run by removing the internal trace-to-run relationships.

        Args:
            location_id: The location ID (e.g., "catalog.schema") for the traces.
            otel_trace_ids: List of OTEL trace IDs to unlink from the run.
            run_id: ID of the run to unlink traces from.
        """
        if not otel_trace_ids:
            return

        req_body = message_to_json(
            BatchUnlinkTraceFromRun(
                location_id=location_id,
                trace_ids=otel_trace_ids,
                run_id=run_id,
            )
        )
        endpoint = f"{_V4_TRACE_REST_API_PATH_PREFIX}/{location_id}/unlink-from-run/batchDelete"
        self._call_endpoint(BatchUnlinkTraceFromRun, req_body, endpoint=endpoint)

    def link_traces_to_run(self, trace_ids: list[str], run_id: str) -> None:
        """
        Link multiple traces to a run by creating trace-to-run relationships.

        Args:
            trace_ids: List of trace IDs to link to the run.
            run_id: ID of the run to link traces to.
        """
        if not trace_ids:
            return

        traces_by_location = self._group_traces_by_location(trace_ids)

        for location_id, batch_trace_ids in traces_by_location.items():
            if location_id is None:
                super().link_traces_to_run(batch_trace_ids, run_id)
            else:
                self._batch_link_traces_to_run(location_id, batch_trace_ids, run_id)

    def unlink_traces_from_run(self, trace_ids: list[str], run_id: str) -> None:
        """
        Unlink multiple traces from a run by removing trace-to-run relationships.

        Args:
            trace_ids: List of trace IDs to unlink from the run.
            run_id: ID of the run to unlink traces from.
        """
        if not trace_ids:
            return

        traces_by_location = self._group_traces_by_location(trace_ids)

        if v3_trace_ids := traces_by_location.pop(None, []):
            raise MlflowException(
                "Unlinking traces from runs is only supported for traces with UC schema "
                f"locations. Unsupported trace IDs: {v3_trace_ids}"
            )

        for location_id, batch_trace_ids in traces_by_location.items():
            self._batch_unlink_traces_from_run(location_id, batch_trace_ids, run_id)

    def _append_sql_warehouse_id_param(self, endpoint: str) -> str:
        if sql_warehouse_id := MLFLOW_TRACING_SQL_WAREHOUSE_ID.get():
            return f"{endpoint}?sql_warehouse_id={sql_warehouse_id}"
        return endpoint



import logging
from typing import Sequence

from opentelemetry.sdk.trace import ReadableSpan

from mlflow.entities.span import Span
from mlflow.environment_variables import MLFLOW_ENABLE_ASYNC_TRACE_LOGGING
from mlflow.tracing.export.async_export_queue import Task
from mlflow.tracing.export.mlflow_v3 import MlflowV3SpanExporter
from mlflow.tracing.utils import get_active_spans_table_name

_logger = logging.getLogger(__name__)


class DatabricksUCTableSpanExporter(MlflowV3SpanExporter):
    """
    An exporter implementation that logs the traces to Databricks Unity Catalog table.
    """

    def __init__(self, tracking_uri: str | None = None) -> None:
        super().__init__(tracking_uri)
        # Override this to False since spans are logged to UC table instead of artifacts.
        self._should_log_spans_to_artifacts = False

        # Track if we've raised an error for span export to avoid raising it multiple times.
        self._has_raised_span_export_error = False

    def _export_spans_incrementally(self, spans: Sequence[ReadableSpan]) -> None:
        """
        Export spans incrementally as they complete.

        Args:
            spans: Sequence of ReadableSpan objects to export.
        """
        location = get_active_spans_table_name()

        if not location:
            # this should not happen since this exporter is only used when a destination
            # is set to UCSchemaLocation
            _logger.debug("No active spans table name found. Skipping span export.")
            return

        _logger.debug(f"exporting spans to uc table: {location}")

        # Wrapping with MLflow span interface for easier downstream handling
        spans = [Span(span) for span in spans]
        if self._should_log_async():
            self._async_queue.put(
                task=Task(
                    handler=self._client.log_spans,
                    args=(location, spans),
                    error_msg="Failed to log spans to the trace server.",
                )
            )
        else:
            try:
                self._client.log_spans(location, spans)
            except Exception as e:
                if self._has_raised_span_export_error:
                    _logger.debug(f"Failed to log spans to the trace server: {e}", exc_info=True)
                else:
                    _logger.warning(f"Failed to log spans to the trace server: {e}")
                    self._has_raised_span_export_error = True

    def _should_enable_async_logging(self) -> bool:
        return MLFLOW_ENABLE_ASYNC_TRACE_LOGGING.get()



import logging

from mlflow.entities.model_registry import ModelVersion, RegisteredModel
from mlflow.entities.webhook import Webhook, WebhookEvent, WebhookStatus, WebhookTestResult
from mlflow.protos.model_registry_pb2 import (
    CreateModelVersion,
    CreateRegisteredModel,
    DeleteModelVersion,
    DeleteModelVersionTag,
    DeleteRegisteredModel,
    DeleteRegisteredModelAlias,
    DeleteRegisteredModelTag,
    GetLatestVersions,
    GetModelVersion,
    GetModelVersionByAlias,
    GetModelVersionDownloadUri,
    GetRegisteredModel,
    ModelRegistryService,
    RenameRegisteredModel,
    SearchModelVersions,
    SearchRegisteredModels,
    SetModelVersionTag,
    SetRegisteredModelAlias,
    SetRegisteredModelTag,
    TransitionModelVersionStage,
    UpdateModelVersion,
    UpdateRegisteredModel,
)
from mlflow.protos.webhooks_pb2 import (
    CreateWebhook,
    DeleteWebhook,
    GetWebhook,
    ListWebhooks,
    TestWebhook,
    UpdateWebhook,
    WebhookService,
)
from mlflow.store.entities.paged_list import PagedList
from mlflow.store.model_registry.base_rest_store import BaseRestStore
from mlflow.utils.proto_json_utils import message_to_json
from mlflow.utils.rest_utils import (
    _REST_API_PATH_PREFIX,
    call_endpoint,
    extract_all_api_info_for_service,
    extract_api_info_for_service,
)

_METHOD_TO_INFO = extract_api_info_for_service(ModelRegistryService, _REST_API_PATH_PREFIX)
_METHOD_TO_ALL_INFO = extract_all_api_info_for_service(ModelRegistryService, _REST_API_PATH_PREFIX)
_WEBHOOK_METHOD_TO_INFO = extract_api_info_for_service(WebhookService, _REST_API_PATH_PREFIX)

_logger = logging.getLogger(__name__)


class RestStore(BaseRestStore):
    """
    Client for a remote model registry server accessed via REST API calls

    Args:
        get_host_creds: Method to be invoked prior to every REST request to get the
            :py:class:`mlflow.rest_utils.MlflowHostCreds` for the request. Note that this
            is a function so that we can obtain fresh credentials in the case of expiry.
    """

    def _get_response_from_method(self, method):
        return method.Response()

    def _get_endpoint_from_method(self, method):
        return _METHOD_TO_INFO[method]

    def _get_all_endpoints_from_method(self, method):
        return _METHOD_TO_ALL_INFO[method]

    def _get_webhook_endpoint_from_method(self, method):
        return _WEBHOOK_METHOD_TO_INFO[method]

    def _call_webhook_endpoint(
        self,
        api,
        json_body: str | None = None,
        webhook_id: str | None = None,
    ):
        endpoint, method = self._get_webhook_endpoint_from_method(api)
        if webhook_id:
            endpoint = endpoint.format(webhook_id=webhook_id)
        response_proto = self._get_response_from_method(api)
        return call_endpoint(self.get_host_creds(), endpoint, method, json_body, response_proto)

    # CRUD API for RegisteredModel objects

    def create_registered_model(self, name, tags=None, description=None, deployment_job_id=None):
        """
        Create a new registered model in backend store.

        Args:
            name: Name of the new model. This is expected to be unique in the backend store.
            tags: A list of :py:class:`mlflow.entities.model_registry.RegisteredModelTag`
                instances associated with this registered model.
            description: Description of the model.
            deployment_job_id: Optional deployment job ID.

        Returns:
            A single object of :py:class:`mlflow.entities.model_registry.RegisteredModel`
            created in the backend.
        """
        proto_tags = [tag.to_proto() for tag in tags or []]
        req_body = message_to_json(
            CreateRegisteredModel(name=name, tags=proto_tags, description=description)
        )
        response_proto = self._call_endpoint(CreateRegisteredModel, req_body)
        return RegisteredModel.from_proto(response_proto.registered_model)

    def update_registered_model(self, name, description, deployment_job_id=None):
        """
        Update description of the registered model.

        Args:
            name: Registered model name.
            description: New description.
            deployment_job_id: Optional deployment job ID.

        Returns:
            A single updated :py:class:`mlflow.entities.model_registry.RegisteredModel` object.
        """
        req_body = message_to_json(UpdateRegisteredModel(name=name, description=description))
        response_proto = self._call_endpoint(UpdateRegisteredModel, req_body)
        return RegisteredModel.from_proto(response_proto.registered_model)

    def rename_registered_model(self, name, new_name):
        """
        Rename the registered model.

        Args:
            name: Registered model name.
            new_name: New proposed name.

        Returns:
            A single updated :py:class:`mlflow.entities.model_registry.RegisteredModel` object.

        """
        req_body = message_to_json(RenameRegisteredModel(name=name, new_name=new_name))
        response_proto = self._call_endpoint(RenameRegisteredModel, req_body)
        return RegisteredModel.from_proto(response_proto.registered_model)

    def delete_registered_model(self, name):
        """
        Delete the registered model.
        Backend raises exception if a registered model with given name does not exist.

        Args:
            name: Registered model name.

        Returns:
            None
        """
        req_body = message_to_json(DeleteRegisteredModel(name=name))
        self._call_endpoint(DeleteRegisteredModel, req_body)

    def search_registered_models(
        self, filter_string=None, max_results=None, order_by=None, page_token=None
    ):
        """
        Search for registered models in backend that satisfy the filter criteria.

        Args:
            filter_string: Filter query string, defaults to searching all registered models.
            max_results: Maximum number of registered models desired.
            order_by: List of column names with ASC|DESC annotation, to be used for ordering
                matching search results.
            page_token: Token specifying the next page of results. It should be obtained from
                a ``search_registered_models`` call.

        Returns:
            A PagedList of :py:class:`mlflow.entities.model_registry.RegisteredModel` objects
            that satisfy the search expressions. The pagination token for the next page can be
            obtained via the ``token`` attribute of the object.

        """
        req_body = message_to_json(
            SearchRegisteredModels(
                filter=filter_string,
                max_results=max_results,
                order_by=order_by,
                page_token=page_token,
            )
        )
        response_proto = self._call_endpoint(SearchRegisteredModels, req_body)
        registered_models = [
            RegisteredModel.from_proto(registered_model)
            for registered_model in response_proto.registered_models
        ]
        return PagedList(registered_models, response_proto.next_page_token)

    def get_registered_model(self, name):
        """
        Get registered model instance by name.

        Args:
            name: Registered model name.

        Returns:
            A single :py:class:`mlflow.entities.model_registry.RegisteredModel` object.
        """
        req_body = message_to_json(GetRegisteredModel(name=name))
        response_proto = self._call_endpoint(GetRegisteredModel, req_body)
        return RegisteredModel.from_proto(response_proto.registered_model)

    def get_latest_versions(self, name, stages=None):
        """
        Latest version models for each requested stage. If no ``stages`` argument is provided,
        returns the latest version for each stage.

        Args:
            name: Registered model name.
            stages: List of desired stages. If input list is None, return latest versions for
                each stage.

        Returns:
            List of :py:class:`mlflow.entities.model_registry.ModelVersion` objects.
        """
        req_body = message_to_json(GetLatestVersions(name=name, stages=stages))
        response_proto = self._call_endpoint(GetLatestVersions, req_body, call_all_endpoints=True)
        return [
            ModelVersion.from_proto(model_version)
            for model_version in response_proto.model_versions
        ]

    def set_registered_model_tag(self, name, tag):
        """
        Set a tag for the registered model.

        Args:
            name: Registered model name.
            tag: :py:class:`mlflow.entities.model_registry.RegisteredModelTag` instance to log.

        Returns:
            None
        """
        req_body = message_to_json(SetRegisteredModelTag(name=name, key=tag.key, value=tag.value))
        self._call_endpoint(SetRegisteredModelTag, req_body)

    def delete_registered_model_tag(self, name, key):
        """
        Delete a tag associated with the registered model.

        Args:
            name: Registered model name.
            key: Registered model tag key.

        Returns:
            None
        """
        req_body = message_to_json(DeleteRegisteredModelTag(name=name, key=key))
        self._call_endpoint(DeleteRegisteredModelTag, req_body)

    # CRUD API for ModelVersion objects

    def create_model_version(
        self,
        name,
        source,
        run_id=None,
        tags=None,
        run_link=None,
        description=None,
        local_model_path=None,
        model_id: str | None = None,
    ):
        """
        Create a new model version from given source and run ID.

        Args:
            name: Registered model name.
            source: URI indicating the location of the model artifacts.
            run_id: Run ID from MLflow tracking server that generated the model.
            tags: A list of :py:class:`mlflow.entities.model_registry.ModelVersionTag`
                instances associated with this model version.
            run_link: Link to the run from an MLflow tracking server that generated this model.
            description: Description of the version.
            local_model_path: Unused.
            model_id: The ID of the model (from an Experiment) that is being promoted to a
                registered model version, if applicable.

        Returns:
            A single object of :py:class:`mlflow.entities.model_registry.ModelVersion`
            created in the backend.

        """
        proto_tags = [tag.to_proto() for tag in tags or []]
        req_body = message_to_json(
            CreateModelVersion(
                name=name,
                source=source,
                run_id=run_id,
                run_link=run_link,
                tags=proto_tags,
                description=description,
                model_id=model_id,
            )
        )
        response_proto = self._call_endpoint(CreateModelVersion, req_body)
        return ModelVersion.from_proto(response_proto.model_version)

    def transition_model_version_stage(self, name, version, stage, archive_existing_versions):
        """
        Update model version stage.

        Args:
            name: Registered model name.
            version: Registered model version.
            stage: New desired stage for this model version.
            archive_existing_versions: If this flag is set to ``True``, all existing model
                versions in the stage will be automatically moved to the "archived" stage. Only
                valid when ``stage`` is ``"staging"`` or ``"production"`` otherwise an error will
                be raised.

        Returns:
            A single :py:class:`mlflow.entities.model_registry.ModelVersion` object.

        """
        req_body = message_to_json(
            TransitionModelVersionStage(
                name=name,
                version=str(version),
                stage=stage,
                archive_existing_versions=archive_existing_versions,
            )
        )
        response_proto = self._call_endpoint(TransitionModelVersionStage, req_body)
        return ModelVersion.from_proto(response_proto.model_version)

    def update_model_version(self, name, version, description):
        """
        Update metadata associated with a model version in backend.

        Args:
            name: Registered model name.
            version: Registered model version.
            description: New model description.

        Returns:
            A single :py:class:`mlflow.entities.model_registry.ModelVersion` object.

        """
        req_body = message_to_json(
            UpdateModelVersion(name=name, version=str(version), description=description)
        )
        response_proto = self._call_endpoint(UpdateModelVersion, req_body)
        return ModelVersion.from_proto(response_proto.model_version)

    def delete_model_version(self, name, version):
        """
        Delete model version in backend.

        Args:
            name: Registered model name.
            version: Registered model version.

        Returns:
            None
        """
        req_body = message_to_json(DeleteModelVersion(name=name, version=str(version)))
        self._call_endpoint(DeleteModelVersion, req_body)

    def get_model_version(self, name, version):
        """
        Get the model version instance by name and version.

        Args:
            name: Registered model name.
            version: Registered model version.

        Returns:
            A single :py:class:`mlflow.entities.model_registry.ModelVersion` object.
        """
        req_body = message_to_json(GetModelVersion(name=name, version=str(version)))
        response_proto = self._call_endpoint(GetModelVersion, req_body)
        return ModelVersion.from_proto(response_proto.model_version)

    def get_model_version_download_uri(self, name, version):
        """
        Get the download location in Model Registry for this model version.
        NOTE: For first version of Model Registry, since the models are not copied over to another
              location, download URI points to input source path.

        Args:
            name: Registered model name.
            version: Registered model version.

        Returns:
            A single URI location that allows reads for downloading.
        """
        req_body = message_to_json(GetModelVersionDownloadUri(name=name, version=str(version)))
        response_proto = self._call_endpoint(GetModelVersionDownloadUri, req_body)
        return response_proto.artifact_uri

    def search_model_versions(
        self, filter_string=None, max_results=None, order_by=None, page_token=None
    ):
        """
        Search for model versions in backend that satisfy the filter criteria.

        Args:
            filter_string: A filter string expression. Currently supports a single filter
                condition either name of model like ``name = 'model_name'`` or
                ``run_id = '...'``.
            max_results: Maximum number of model versions desired.
            order_by: List of column names with ASC|DESC annotation, to be used for ordering
                matching search results.
            page_token: Token specifying the next page of results. It should be obtained from
                a ``search_model_versions`` call.

        Returns:
            A PagedList of :py:class:`mlflow.entities.model_registry.ModelVersion`
            objects that satisfy the search expressions. The pagination token for the next
            page can be obtained via the ``token`` attribute of the object.

        """
        req_body = message_to_json(
            SearchModelVersions(
                filter=filter_string,
                max_results=max_results,
                order_by=order_by,
                page_token=page_token,
            )
        )
        response_proto = self._call_endpoint(SearchModelVersions, req_body)
        model_versions = [ModelVersion.from_proto(mvd) for mvd in response_proto.model_versions]
        return PagedList(model_versions, response_proto.next_page_token)

    def set_model_version_tag(self, name, version, tag):
        """
        Set a tag for the model version.

        Args:
            name: Registered model name.
            version: Registered model version.
            tag: :py:class:`mlflow.entities.model_registry.ModelVersionTag` instance to log.

        Returns:
            None
        """
        req_body = message_to_json(
            SetModelVersionTag(name=name, version=str(version), key=tag.key, value=tag.value)
        )
        self._call_endpoint(SetModelVersionTag, req_body)

    def delete_model_version_tag(self, name, version, key):
        """
        Delete a tag associated with the model version.

        Args:
            name: Registered model name.
            version: Registered model version.
            key: Tag key.

        Returns:
            None
        """
        req_body = message_to_json(DeleteModelVersionTag(name=name, version=str(version), key=key))
        self._call_endpoint(DeleteModelVersionTag, req_body)

    def set_registered_model_alias(self, name, alias, version):
        """
        Set a registered model alias pointing to a model version.

        Args:
            name: Registered model name.
            alias: Name of the alias.
            version: Registered model version number.

        Returns:
            None
        """
        req_body = message_to_json(
            SetRegisteredModelAlias(name=name, alias=alias, version=str(version))
        )
        self._call_endpoint(SetRegisteredModelAlias, req_body)

    def delete_registered_model_alias(self, name, alias):
        """
        Delete an alias associated with a registered model.

        Args:
            name: Registered model name.
            alias: Name of the alias.

        Returns:
            None
        """
        req_body = message_to_json(DeleteRegisteredModelAlias(name=name, alias=alias))
        self._call_endpoint(DeleteRegisteredModelAlias, req_body)

    def get_model_version_by_alias(self, name, alias):
        """
        Get the model version instance by name and alias.

        Args:
            name: Registered model name.
            alias: Name of the alias.

        Returns:
            A single :py:class:`mlflow.entities.model_registry.ModelVersion` object.
        """
        req_body = message_to_json(GetModelVersionByAlias(name=name, alias=alias))
        response_proto = self._call_endpoint(GetModelVersionByAlias, req_body)
        return ModelVersion.from_proto(response_proto.model_version)

    # Webhook APIs
    def create_webhook(
        self,
        name: str,
        url: str,
        events: list[WebhookEvent],
        description: str | None = None,
        secret: str | None = None,
        status: WebhookStatus | None = None,
    ) -> Webhook:
        req_body = message_to_json(
            CreateWebhook(
                name=name,
                url=url,
                events=[e.to_proto() for e in events],
                description=description,
                secret=secret,
                status=status.to_proto() if status else None,
            )
        )
        response_proto = self._call_webhook_endpoint(CreateWebhook, req_body)
        return Webhook.from_proto(response_proto.webhook)

    def get_webhook(self, webhook_id: str) -> Webhook:
        response_proto = self._call_webhook_endpoint(GetWebhook, webhook_id=webhook_id)
        return Webhook.from_proto(response_proto.webhook)

    def list_webhooks(
        self,
        max_results: int | None = None,
        page_token: str | None = None,
    ) -> PagedList[Webhook]:
        req_body = message_to_json(ListWebhooks(max_results=max_results, page_token=page_token))
        response_proto = self._call_webhook_endpoint(ListWebhooks, req_body)
        webhooks = [Webhook.from_proto(webhook) for webhook in response_proto.webhooks]
        return PagedList(webhooks, response_proto.next_page_token)

    def update_webhook(
        self,
        webhook_id: str,
        name: str | None = None,
        description: str | None = None,
        url: str | None = None,
        events: list[WebhookEvent] | None = None,
        secret: str | None = None,
        status: WebhookStatus | None = None,
    ) -> Webhook:
        req_body = message_to_json(
            UpdateWebhook(
                name=name,
                description=description,
                url=url,
                events=[e.to_proto() for e in events] if events else None,
                secret=secret,
                status=status.to_proto() if status else None,
            )
        )
        response_proto = self._call_webhook_endpoint(UpdateWebhook, req_body, webhook_id=webhook_id)
        return Webhook.from_proto(response_proto.webhook)

    def delete_webhook(self, webhook_id: str) -> None:
        self._call_webhook_endpoint(DeleteWebhook, webhook_id=webhook_id)

    def test_webhook(self, webhook_id: str, event: WebhookEvent | None = None) -> WebhookTestResult:
        """
        Test the webhook by sending a test event to the specified URL.

        Args:
            webhook_id: The ID of the webhook to test.
            event: Optional event type to test. If not specified, uses the first event from webhook.

        Returns:
            WebhookTestResult indicating success/failure and response details
        """
        req_body = message_to_json(TestWebhook(event=event.to_proto() if event else None))
        response_proto = self._call_webhook_endpoint(TestWebhook, req_body, webhook_id=webhook_id)
        return WebhookTestResult.from_proto(response_proto.result)
